from datetime import timezone
from enum import StrEnum
from typing import Annotated

from pydantic import AfterValidator, AwareDatetime


def to_utc(dt: AwareDatetime) -> AwareDatetime:
    return dt.astimezone(timezone.utc)


DatetimeUTC = Annotated[AwareDatetime, AfterValidator(to_utc)]


class PhoneNumberType(StrEnum):
    """all possible types for a phone number"""

    PERSONAL = "personal"
    FUNCTIONAL = "functional"
    EXTERNAL = "external"
    BLOCKED = "blocked"
    LEGACY = "legacy"



"""Personal number history schema"""

from typing import override

from zih_models.tkdb.types import DatetimeUTC

from .base import (
    BaseDeleteSchemaModel,
    BaseDiffSchemaModel,
    BaseSchemaModel,
    FieldDiff,
    Login,
    TelephoneNumber,
)

"""
!!! NEVER Change the type of a Column/Attribute of this models so the history never breaks !!!
"""


class PersonalNumber(BaseSchemaModel):

    telephone_number: TelephoneNumber
    user_common_name: Login
    primary: bool
    voicemail: bool
    unused_since: DatetimeUTC | None
    comment: str | None
    dod: str | None

    table = "personal_numbers"

    @override
    def get_user_logins(self) -> set[str]:
        logins = super().get_user_logins()
        logins.add(self.user_common_name)
        return logins


class PersonalNumberDiff(BaseDiffSchemaModel):

    primary: FieldDiff[bool] = None
    voicemail: FieldDiff[bool] = None
    unused_since: FieldDiff[DatetimeUTC | None] = None
    comment: FieldDiff[str | None] = None
    dod: FieldDiff[str | None] = None

    table = "personal_numbers"


class PersonalNumberDelete(BaseDeleteSchemaModel):
    """delete model"""

    table = "personal_numbers"



"""models for api"""

from enum import StrEnum
from typing import Annotated, Any, Self

from pydantic import BaseModel, Field, model_validator

from zih_models.tkdb.types import DatetimeUTC, PhoneNumberType

from .history_models.base import TABLE, BaseSchemaModel, ChangeAction

TelephoneNumber = Annotated[
    str,
    Field(
        pattern=r"^\+49-\d{1,3}-\d{0,8}-\d{1,12}$",
        examples=[
            "+49-351-463-40000",
            "+49-351-463-50000",
            "+49-351-463-60000",
        ],
    ),
]


class TelephoneNumberAssignment(BaseModel):
    """Base class of all Phone number assignment classes."""

    telephone_number: TelephoneNumber


class FunctionalPhoneNumber(TelephoneNumberAssignment):
    """FunctionalPhone with base information"""

    organizational_unit: str


class FunctionalPhoneNumberWithDetails(FunctionalPhoneNumber):
    """FunctionalPhone with all information for a detailed view"""

    manager_cn: str
    voicemail_user: list[str]
    comment: str | None = None
    category_id: int | None = None
    dod_id: int | None
    created: DatetimeUTC
    last_modified: DatetimeUTC
    unused_since: DatetimeUTC | None


class FunctionalPhoneNumberWrite(FunctionalPhoneNumber):
    """FunctionalPhone with all optional write arguments"""

    voicemail_user: list[str] | None = None
    comment: str | None = None
    category_id: int | None = None
    dod_id: int


class PersonalPhoneNumber(TelephoneNumberAssignment):
    pass


class PersonalPhoneNumberWithDetails(PersonalPhoneNumber):
    user_cn: str
    comment: str | None = None
    dod_id: int | None
    voicemail: bool
    created: DatetimeUTC
    last_modified: DatetimeUTC
    unused_since: DatetimeUTC | None


class PersonalPhoneNumberWrite(PersonalPhoneNumber):
    comment: str | None
    voicemail: bool = False
    dod_id: int


class ExternalPhoneNumberWithDetails(TelephoneNumberAssignment):
    organization_id: int
    dod_id: int | None
    comment: str | None = None
    created: DatetimeUTC
    last_modified: DatetimeUTC
    unused_since: DatetimeUTC | None


class UserAssignmentsBase[
    TIpPhone: (str, PersonalPhoneNumberWrite, PersonalPhoneNumberWithDetails),
    TManagerOfPhone: (
        FunctionalPhoneNumberWrite,
        FunctionalPhoneNumber,
        FunctionalPhoneNumberWithDetails,
    ),
](BaseModel):
    """UserAssignments base class"""

    ipPhone: TIpPhone | None
    otherIpPhone: list[TIpPhone]
    managerOfPhone: list[TManagerOfPhone]


class UserAssignmentsWrite(
    UserAssignmentsBase[PersonalPhoneNumberWrite, FunctionalPhoneNumberWrite]
):
    """UserAssignments without common_name to write to api and more validation"""

    @model_validator(mode="after")
    def validate_primary(self) -> Self:
        """validate that ipPhone is required if you want to set otherIpPhones"""
        if self.otherIpPhone and self.ipPhone is None:
            raise ValueError(
                "ipPhone is required if you want to set otherIpPhones"
            )
        return self

    @model_validator(mode="after")
    def check_unique_telephone_number_across_attributes(self) -> Self:
        """validate that ipPhone, otherIpPhones and managerOfPhone dont have duplicated telephone_numbers"""
        personal_numbers: list[
            PersonalPhoneNumberWrite | FunctionalPhoneNumberWrite
        ] = ([self.ipPhone, *self.otherIpPhone] if self.ipPhone else [])
        all_numbers = [
            item.telephone_number
            for item in personal_numbers + self.managerOfPhone
        ]
        if len(all_numbers) != len(set(all_numbers)):
            raise ValueError(
                "Telephone numbers must be unique across ipPhone, otherIpPhones and managerOfPhone."
            )
        return self


class UserAssignments(UserAssignmentsBase[str, FunctionalPhoneNumber]):
    """UserAssignments return for api"""

    common_name: str

    @property
    def telephone_numbers(self) -> list[str]:
        """return all telephone numbers of user. (no numbers where the user is manager)"""
        return (
            [
                self.ipPhone,
                *self.otherIpPhone,
            ]
            if self.ipPhone is not None
            else self.otherIpPhone
        )


class UserAssignmentsWithDetails(
    UserAssignmentsBase[
        PersonalPhoneNumberWithDetails, FunctionalPhoneNumberWithDetails
    ]
):
    """UserAssignments return for api"""

    common_name: str


class HistoryEntry[
    T: (dict[str, Any], BaseSchemaModel) = dict[str, Any],
    TAction: ChangeAction = ChangeAction,
](BaseModel):

    subject: str | None
    source: str
    timestamp: DatetimeUTC
    entity_type: TABLE
    entity_identifier: str
    action: TAction
    data: T


class NewPrimaryBody(BaseModel):

    telephone_number: TelephoneNumber


class LegacyTKDBLink(BaseModel):
    """Object relation for an object that is currently not on the ssp display and is a link to old tkdb"""

    link: str
    display: str


class GatewayPort(BaseModel):
    gateway: LegacyTKDBLink
    port: int
    room: Annotated[
        str,
        Field(
            pattern=r"^TU-[A-Z]{2}-\d{3}-\d{4}-\d{2}-\d{4}$",
            examples=[
                "TU-DD-145-1231-00-5830",
            ],
        ),
    ]
    comment: str
    extra_comment: str


class IsdnCable(BaseModel):
    name: LegacyTKDBLink
    comment: str
    extra_comment: str


class VoipPhone(BaseModel):
    device: LegacyTKDBLink
    line: int


class PhoneNumberDetails(BaseModel):

    id: int
    number: str
    prefix: str
    dial_up: str
    extension: str
    type: PhoneNumberType
    use_id: int | None
    comment: str | None
    assignment: str | None
    partition: str | None
    isdn_system: int | None
    isdn_port: str | None

    # combined values
    # is set on personal and function numbers / also used for type legacy
    voicemail_users: list[str] | None = None

    # If the number has any active usage, this is False. Is None for blocked numbers.
    unused: bool | None = None

    # external number values / also used for type legacy
    external_organization_id: int | None = None

    # functional number values / also used for type legacy
    organization_unit: str | None = None
    manager: str | None = None
    functional_category: int | None = None

    # personal number values
    # for better implementation this None on all legacy numbers, even if they have an personal relation
    user: str | None = None

    functional_details: FunctionalPhoneNumberWithDetails | None = None
    personal_details: list[PersonalPhoneNumberWithDetails] | None = None
    external_details: ExternalPhoneNumberWithDetails | None = None

    gateway_ports: list[GatewayPort] | None = None
    isdn_cables: list[IsdnCable] | None = None
    voip_phones: list[VoipPhone] | None = None


class PhoneNumbersResponse(BaseModel):

    items: list[PhoneNumberDetails]
    total: int


class AllowedTelephoneFilterFields(StrEnum):

    ID = "id"
    NUMBER = "number"
    PREFIX = "prefix"
    DIAL_UP = "dial_up"
    EXTENSION = "extension"
    TYPE = "type"
    USE_ID = "use_id"
    COMMENT = "comment"
    ASSIGNMENT = "assignment"
    PARTITION = "partition"
    ISDN_SYSTEM = "isdn_system"
    ISDN_PORT = "isdn_port"
    VOICEMAIL_USERS = "voicemail_users"
    UNUSED = "unused"
    EXTERNAL_ORGANIZATION_ID = "external_organization_id"
    ORGANIZATION_UNIT = "organization_unit"
    MANAGER = "manager"
    FUNCTIONAL_CATEGORY = "functional_category"
    USER = "user"


class StatsResponse(BaseModel):

    # used states of automatic number ranges
    free_in_number_ranges: int
    used_in_number_ranges: int

    # type counts
    personal_numbers: int
    blocked_numbers: int
    external_numbers: int
    functional_numbers: int
    legacy_numbers: int

    # legacy count splitted in legacy reasons
    legacy_reasons: dict[str, int]

    # use categories
    isdn_numbers: int
    voip_numbers: int



"""External number history schema"""

from zih_models.tkdb.types import DatetimeUTC

from .base import (
    BaseDeleteSchemaModel,
    BaseDiffSchemaModel,
    BaseSchemaModel,
    FieldDiff,
    TelephoneNumber,
)

"""
!!! NEVER Change the type of a Column/Attribute of this models so the history never breaks !!!
"""


class ExternalNumber(BaseSchemaModel):

    telephone_number: TelephoneNumber
    organization: str
    unused_since: DatetimeUTC | None
    comment: str | None
    dod: str | None

    table = "external_numbers"


class ExternalNumberDiff(BaseDiffSchemaModel):

    unused_since: FieldDiff[DatetimeUTC | None] = None
    comment: FieldDiff[str | None] = None
    dod: FieldDiff[str | None] = None

    table = "external_numbers"


class ExternalNumberDelete(BaseDeleteSchemaModel):
    """delete model"""

    table = "external_numbers"



"""Functional number history schema"""

from typing import override

from zih_models.tkdb.types import DatetimeUTC

from .base import (
    OU,
    BaseDeleteSchemaModel,
    BaseDiffSchemaModel,
    BaseSchemaModel,
    FieldDiff,
    Login,
    TelephoneNumber,
)

"""
!!! NEVER Change the type of a Column/Attribute of this models so the history never breaks !!!
"""


class FunctionalNumber(BaseSchemaModel):

    telephone_number: TelephoneNumber
    manager_uid: Login
    organizational_unit: OU
    voicemail_user: list[Login]
    unused_since: DatetimeUTC | None
    category: str | None
    comment: str | None
    dod: str | None

    table = "functional_numbers"

    @override
    def get_user_logins(self) -> set[str]:
        logins = super().get_user_logins()
        logins.update(self.manager_uid, *self.voicemail_user)
        return logins

    @override
    def get_ous(self) -> set[str]:
        ous = super().get_ous()
        ous.add(self.organizational_unit)
        return ous


class FunctionalNumberDiff(BaseDiffSchemaModel):

    manager_uid: FieldDiff[Login] = None
    organizational_unit: FieldDiff[OU] = None
    voicemail_user: FieldDiff[list[Login]] = None
    unused_since: FieldDiff[DatetimeUTC | None] = None
    category: FieldDiff[str | None] = None
    comment: FieldDiff[str | None] = None
    dod: FieldDiff[str | None] = None

    table = "functional_numbers"

    @override
    def get_user_logins(self) -> set[str]:
        logins = super().get_user_logins()
        if self.voicemail_user:
            logins.update(*self.voicemail_user.old, *self.voicemail_user.new)
        if self.manager_uid:
            logins.update(self.manager_uid.old, self.manager_uid.new)
        return logins

    @override
    def get_ous(self) -> set[str]:
        ous = super().get_ous()
        if self.organizational_unit:
            ous.update(
                self.organizational_unit.old, self.organizational_unit.new
            )
        return ous


class FunctionalNumberDelete(BaseDeleteSchemaModel):
    """delete model"""

    table = "functional_numbers"



"""User history schema"""

from typing import override

from .base import (
    BaseDeleteSchemaModel,
    BaseDiffSchemaModel,
    BaseSchemaModel,
    FieldDiff,
    FunctionalNumber,
    Login,
    PersonalNumber,
)

"""
!!! NEVER Change the type of a Column/Attribute of this models so the history never breaks !!!
"""


class User(BaseSchemaModel):

    uid: Login
    personal_numbers: list[PersonalNumber]
    functional_numbers: list[FunctionalNumber]

    table = "users_v2"

    @override
    def get_user_logins(self) -> set[str]:
        logins = super().get_user_logins()
        logins.add(self.uid)
        return logins


class UserDiff(BaseDiffSchemaModel):

    personal_numbers: FieldDiff[list[PersonalNumber]] = None
    functional_numbers: FieldDiff[list[FunctionalNumber]] = None

    table = "users_v2"


class UserDelete(BaseDeleteSchemaModel):
    """delete model"""

    table = "users_v2"



# SPDX-License-Identifier: MIT

"""models for the sharepoint api"""

from enum import IntEnum, StrEnum
from typing import Annotated, Literal

from pydantic import BaseModel, Field


class SharepointNotFoundError(BaseModel):
    """Return of the api if target not found"""

    detail: Literal["Not Found"] = "Not Found"


class SiteType(StrEnum):
    """possible types of a site"""

    DEFAULT = "default"
    PROJECT = "project"


class Site(BaseModel):
    """Sharepoint Site"""

    SiteUrl: str
    SiteId: str
    Admin1: str | None = None
    Admin2: str | None = None


class Language(StrEnum):
    """site language"""

    GERMAN = "de"
    ENGLISH = "en"


class SharepointLanguageCodes(IntEnum):
    """sharepoint language codes"""

    GERMAN = 1031
    ENGLISH = 1033


LanguageMapper = {
    Language.GERMAN: SharepointLanguageCodes.GERMAN,
    Language.ENGLISH: SharepointLanguageCodes.ENGLISH,
}

LanguageMapperLogList = {
    Language.GERMAN: "Deutsch (Deutschland)",
    Language.ENGLISH: "Englisch (USA)",
}


class PrivacyLevel(StrEnum):
    """privacy level of site"""

    NORMAL = "normal"
    HIGH = "high"


class User(BaseModel):
    """User Object with all Infos Sharepoint needs"""

    login: str
    mail: str


class CreateSiteData(BaseModel):
    """Sharepoint Site data which is used to create a new Site"""

    applicant: User
    head: User
    it_referent: User
    ticket_number: str
    org_id: str
    site_name: str
    url_name: Annotated[str, Field(max_length=10)]
    site_type: SiteType
    purpose: str
    site_language: Language
    privacy_level: PrivacyLevel
    admins: Annotated[list[User], Field(min_length=2, max_length=2)]



"""base helper things for all schemas"""

from enum import StrEnum
from typing import Annotated, ClassVar, Literal

from pydantic import BaseModel, Field


class ChangeAction(StrEnum):

    CREATED = "created"
    DELETED = "deleted"
    MODIFIED = "modified"


class FieldDiffObject[T](BaseModel):
    old: T
    new: T


type FieldDiff[T] = FieldDiffObject[T] | None

TABLE = Literal[
    "functional_numbers",
    "phone_numbers_v2",
    "personal_numbers",
    "users_v2",
    "external_numbers",
]


class UIDisplaySpecials(StrEnum):
    USER = "user"
    OU = "ou"
    TELEPHONE_NUMBER_LINK = "telephone_number_link"
    PERSONAL_NUMBER_LINK = "personal_number_link"
    FUNCTIONAL_NUMBER_LINK = "functional_number_link"


Login = Annotated[
    str, Field(json_schema_extra={"ui_display": UIDisplaySpecials.USER})
]
OU = Annotated[
    str, Field(json_schema_extra={"ui_display": UIDisplaySpecials.OU})
]
TelephoneNumber = Annotated[
    str,
    Field(
        json_schema_extra={
            "ui_display": UIDisplaySpecials.TELEPHONE_NUMBER_LINK
        }
    ),
]
PersonalNumber = Annotated[
    str,
    Field(
        json_schema_extra={
            "ui_display": UIDisplaySpecials.PERSONAL_NUMBER_LINK
        }
    ),
]
FunctionalNumber = Annotated[
    str,
    Field(
        json_schema_extra={
            "ui_display": UIDisplaySpecials.FUNCTIONAL_NUMBER_LINK
        }
    ),
]


class BaseSchemaModel(BaseModel):
    """Base model with table classvar"""

    table: ClassVar[TABLE]
    action: ClassVar[ChangeAction] = ChangeAction.CREATED

    def get_user_logins(self) -> set[str]:
        return set()

    def get_ous(self) -> set[str]:
        return set()


class BaseDiffSchemaModel(BaseSchemaModel):
    """base for all diff models"""

    action: ClassVar[ChangeAction] = ChangeAction.MODIFIED

    def has_changes(self) -> bool:
        for field in type(self).model_fields:
            if getattr(self, field) is not None:
                return True
        return False


class BaseDeleteSchemaModel(BaseSchemaModel):
    """base for all delete models"""

    action: ClassVar[ChangeAction] = ChangeAction.DELETED
